"""
# SCAR SIM
[![PyPI version](https://badge.fury.io/py/scar_sim.svg)](https://badge.fury.io/py/scar_sim)
[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)

Supply Chain Adaptation and Resilience Simulator

# Setup

Make sure you have Python 3.11.x (or higher) installed on your system. You can download it [here](https://www.python.org/downloads/).

## Installation

```
pip install scar_sim
```

## Getting Started

- [Docs](https://connor-makowski.github.io/scar_sim/scar_sim.html)
- [Git](https://github.com/connor-makowski/scar_sim)

### Basic Example

```py
from scar_sim.entity import Arc, Node
from scar_sim.order import Order
from scar_sim.simulation import Simulation


simulation = Simulation()

# Create nodes
supplier_0 = simulation.add_object(
    Node(
        processing_min_time=0.8,
        processing_avg_time=1.0,
        processing_sd_time=0.02,
        processing_cashflow_per_unit=-50,
        metadata={
            "loc": "cn_ningbo",
            "otype": "node_supplier",
        },
    )
)
factory_1 = simulation.add_object(
    Node(
        processing_min_time=0.2,
        processing_avg_time=0.4,
        processing_sd_time=0.1,
        processing_cashflow_per_unit=-15,
        metadata={
            "loc": "us_ks_kc",
            "otype": "node_factory",
        },
    )
)

# Create arcs between nodes
arc_0_1 = simulation.add_object(
    Arc(
        origin_node=supplier_0,
        destination_node=factory_1,
        processing_min_time=2.0,
        processing_avg_time=2.0,
        processing_sd_time=0.05,
        processing_cashflow_per_unit=-10,
        metadata={
            "loc": "oc_pa",
            "otype": "arc_ocean",
        },
    )
)

order = simulation.add_object(
    Order(
        origin_node=supplier_0,
        destination_node=factory_1,
        units=1,
        planned_path=simulation.graph.get_optimal_path(
            supplier_0, factory_1, "cashflow"
        ),
    )
)

simulation.add_event(
    time_delta=0.0,
    func=order.start,
)

simulation.run(max_time=10.0)

print(simulation.orders[0].history[-1]) #=>
# {
#     'time': 2.9971,
#     'time_delta': 0.0,
#     'order_id': 3,
#     'current_obj_id': 1,
#     'meta': {
#         'loc': 'us_ks_kc',
#         'otype': 'node_factory',
#         'time': 2
#     },
#     'status': 'completed',
#     'cashflow': 0.0
# }

```

## Development

To avoid extra development overhead, we expect all developers to use a unix based environment (Linux or Mac). If you use Windows, please use WSL2.

For development, we test using Docker so we can lock system deps and swap out python versions easily. However, you can also use a virtual environment if you prefer. We provide a test script and a prettify script to help with development.

## Making Changes

1) Fork the repo and clone it locally.
2) Make your modifications.
3) Use Docker or a virtual environment to run tests and make sure they pass.
4) Prettify your code.
5) **DO NOT GENERATE DOCS**.
    - We will generate the docs and update the version number when we are ready to release a new version.
6) Only commit relevant changes and add clear commit messages.
    - Atomic commits are preferred.
7) Submit a pull request.

## Docker

Make sure Docker is installed and running.

- Create a docker container and drop into a shell
    - `./run.sh`
- Run all tests (see ./utils/test.sh)
    - `./run.sh test`
- Prettify the code (see ./utils/prettify.sh)
    - `./run.sh prettify`

- Note: You can and should modify the `Dockerfile` to test different python versions.

## Virtual Environment

- Create a virtual environment
    - `python3.XX -m venv venv`
        - Replace `3.XX` with your python version (3.11 or higher)
- Activate the virtual environment
    - `source venv/bin/activate`
- Install the development requirements
    - `pip install -r requirements/dev.txt`
- Run Tests
    - `./utils/test.sh`
- Prettify Code
    - `./utils/prettify.sh`"""




from heapq import heappop, heappush
from scar_sim.utils import hard_round


class Queue:
    def __init__(self, log_events: bool = False, precision: int = 4):
        """
        Initializes a priority queue for managing simulation events.

        Optional Arguments:

        - log_events (bool): If True, logs each event processed for debugging or analysis. Default is False.
        - precision (int): The number of decimal places to round time values to, ensuring numerical stability.
            - Default: 4
        """
        self.__queue__ = []
        self.__current_time__ = 0.0
        self.__log__ = []
        self.__event_dict__ = {}
        self.__event_id__ = 0
        self.__log_events__ = log_events
        self.__precision__ = precision

    def add(
        self,
        time_delta: float,
        func,
        args: tuple = tuple(),
        kwargs: dict = dict(),
    ) -> None:
        """
        Schedules a new event in the queue to be executed after a specified time delta.

        Required Arguments:

        - time_delta (float): The time delay after which the event should be executed. Must be non-negative.
        - func (callable): The function to be called when the event is processed.

        Optional Arguments:

        - args (tuple): Positional arguments to pass to the function when called.
            - Default: tuple()
        - kwargs (dict): Keyword arguments to pass to the function when called.
            - Default: dict()
        """
        if time_delta < 0:
            raise ValueError("Cannot schedule events in the past")
        self.__event_id__ += 1
        self.__event_dict__[self.__event_id__] = (
            self.__event_id__,
            func,
            args,
            kwargs,
        )
        next_time = hard_round(
            self.__current_time__ + time_delta, self.__precision__
        )
        heappush(self.__queue__, (next_time, self.__event_id__))

    def process(self) -> None:
        """
        Processes the next event in the queue, updating the current time and executing the associated function.

        Logs the event if logging is enabled.

        Raises:

        - IndexError: If there are no events to process in the queue.

        Returns:

        - None
        """
        if not self.__queue__:
            raise IndexError("No events to process in the queue")
        self.__current_time__, event_id = heappop(self.__queue__)
        event_id, func, args, kwargs = self.__event_dict__.pop(event_id)
        func(*args, **kwargs)

        if self.__log_events__:
            self.__log__.append(
                {
                    "event_id": event_id,
                    "time": self.__current_time__,
                    "func": func.__name__,
                    "args": args,
                    "kwargs": kwargs,
                }
            )

    def run(self, max_time: float) -> None:
        """
        Runs the event queue, processing events in chronological order until the specified maximum time is reached.

        Once the simulation is complete, the current time is set to max_time to ensure future events are scheduled correctly.

        Required Arguments:

        - max_time (float): The maximum simulation time at which to stop processing events.

        Raises:

        - ValueError: If max_time is less than the current simulation time.

        Returns:

        - None
        """
        if max_time < self.__current_time__:
            raise ValueError(
                "max_time cannot be less than the current simulation time"
            )
        while self.__queue__ and self.__queue__[0][0] <= max_time:
            self.process()
        self.__current_time__ = max_time



import random

random.seed(42)


class SimulationObject:
    def __init__(self):
        """
        Initializes a SimulationObject with default attributes.

        This object forms the base class for all entities within the simulation framework.
        """
        self.id = None
        """ A unique identifier for the simulation object. Initially set to None."""
        self.__simulation__ = None
        """ A reference to the simulation instance that this object belongs to. Initially set to None."""


class SimulationEntity(SimulationObject):
    def __init__(
        self,
        processing_min_time: float = 0.0,
        processing_avg_time: float = 0.0,
        processing_sd_time: float = 0.0,
        processing_cashflow_per_unit: float = 0.0,
        metadata: dict = dict(),
    ):
        """
        Initializes a SimulationEntity with processing parameters and metadata.

        Required Arguments:

        - processing_min_time (float): The minimum processing time for the entity.
        - processing_avg_time (float): The average processing time for the entity.
        - processing_sd_time (float): The standard deviation of processing time for the entity.
        - processing_cashflow_per_unit (float): The cashflow per unit processed by the entity

        Optional Arguments:

        - metadata (dict): A dictionary of metadata associated with the entity.
            - This is injected into orders and can be used for tracking and analysis.
        """
        # Basic info
        super().__init__()
        self.entity_type = self.__class__.__name__
        # Split the entity type on uppercase letters and join with underscores
        self.cashflow_key = (
            "".join(
                [
                    "_" + letter.lower() if letter.isupper() else letter
                    for letter in self.entity_type
                ]
            ).lstrip("_")
            + "_cashflow"
        )
        self.metadata = metadata

        # Processing defaults
        self.__default_processing_min_time__ = processing_min_time
        self.__default_processing_avg_time__ = processing_avg_time
        self.__default_processing_sd_time__ = processing_sd_time
        self.__default_processing_cashflow_per_unit__ = (
            processing_cashflow_per_unit
        )

        # Live processing info
        self.processing_time_min = processing_min_time
        self.processing_time_avg = processing_avg_time
        self.processing_time_sd = processing_sd_time
        self.processing_cashflow_per_unit = processing_cashflow_per_unit

    def get_metadata(self, **kwargs) -> dict:
        """
        Retrieve the metadata dictionary for the entity, optionally augmented with additional key-value pairs.

        This method is deisnged to be useful by default but can be overridden in subclasses to provide custom metadata.

        Optional Arguments:

        - kwargs: Additional key-value pairs to include in the returned metadata dictionary.

        Returns:

        - dict: A dictionary containing the entity's metadata, augmented with any additional key-value pairs provided to this method.
        """
        # Return a copy of the metadata dictionary (in case it is modified later)
        return {**dict(self.metadata), **kwargs}

    def get_processing_time(self) -> float:
        """
        Calculates a random processing time based on the entity's processing parameters.

        This method can be overridden in subclasses to provide custom processing time calculations.

        Returns:

        - float: A processing time value, which is at least the minimum processing time and follows a normal distribution defined by the average and standard deviation.
        """
        return max(
            self.processing_time_min,
            random.gauss(self.processing_time_avg, self.processing_time_sd),
        )

    def change_processing_parameters(
        self,
        processing_min_time: float | None = None,
        processing_avg_time: float | None = None,
        processing_sd_time: float | None = None,
        processing_cashflow_per_unit: float | None = None,
    ) -> None:
        """
        Change the processing parameters of the entity.

        This would normally be expected to be scheduled as an event in the simulation to trigger something like a future disruption, maintenance, or upgrade.

        It updates the entity itself as well as the simulation graph if the entity is part of a simulation.

        Optional Arguments:

        - processing_min_time (float | None): New minimum processing time.
        - processing_avg_time (float | None): New average processing time.
        - processing_sd_time (float | None): New standard deviation of processing time.
        - processing_cashflow_per_unit (float | None): New cashflow per unit processed.

        Returns:

        - None
        """
        if processing_min_time is not None:
            self.processing_time_min = processing_min_time
        if processing_avg_time is not None:
            self.processing_time_avg = processing_avg_time
        if processing_sd_time is not None:
            self.processing_time_sd = processing_sd_time
        if processing_cashflow_per_unit is not None:
            self.processing_cashflow_per_unit = processing_cashflow_per_unit
        if self.__simulation__ is not None:
            self.__simulation__.graph.update_graphs(self)

    def reset_processing_parameters(self) -> None:
        """
        Resets the processing parameters of the entity to their default values.

        This would normally be expected to be scheduled as an event in the simulation to trigger the end of a disruption, maintenance, or upgrade.

        This method updates the entity itself as well as the simulation graph if the entity is part of a simulation.

        Returns:

        - None
        """
        self.change_processing_parameters(
            processing_min_time=self.__default_processing_min_time__,
            processing_avg_time=self.__default_processing_avg_time__,
            processing_sd_time=self.__default_processing_sd_time__,
            processing_cashflow_per_unit=self.__default_processing_cashflow_per_unit__,
        )

    def get_cashflow(self, units: int) -> float:
        """
        Calculate the cashflow for processing a given number of units through this entity.

        This method is designed to be useful by default but can be overridden in subclasses to provide custom cashflow calculations.

        Required Arguments:

        - units (int): The number of units being processed.
        """
        return self.processing_cashflow_per_unit * units


class Node(SimulationEntity):
    def __init__(self, **kwargs):
        """
        Initializes a Node entity with processing parameters and metadata.

        Nodes represent points in the simulation where activity occurs, such as facilities or junctions.

        In the underlying graph representation, each node is represented by two nodes (inbound and outbound) connected by an arc that models processing time and cashflow.

        Two additional attributes are initialized to None:

        - inbound_graph_id: The graph ID for the inbound representation of the node.
        - outbound_graph_id: The graph ID for the outbound representation of the node.

        Required Arguments:

        - kwargs: Keyword arguments passed to the SimulationEntity __init__ function.
        """
        super().__init__(**kwargs)
        # Technically each node is made up of two nodes in the graph
        # Connected by an arc that represents processing time and cashflow at the node
        self.inbound_graph_id = None
        self.outbound_graph_id = None

    def order_arrived(self, order):
        """
        A placeholder method to handle logic when an order arrives at this node.
        This can be overridden in subclasses to implement specific behavior.

        Requires:

        - order (Order): The order that has arrived.

        Returns:

        - None
        """

    def order_placed(self, order):
        """
        A placeholder method to handle logic when an order is placed at this node.
        This can be overridden in subclasses to implement specific behavior.

        Requires:

        - order (Order): The order that has been placed.

        Returns:

        - None
        """

    def order_shipped(self, order):
        """
        A placeholder method to handle logic when an order is shipped from this node.
        This can be overridden in subclasses to implement specific behavior.

        Requires:

        - order (Order): The order that has been shipped.

        Returns:

        - None
        """

    def order_completed(self, order):
        """
        A placeholder method to handle logic when an order is completed at this node.
        This can be overridden in subclasses to implement specific behavior.

        Requires:

        - order (Order): The order that has been completed.

        Returns:

        - None
        """


class Arc(SimulationEntity):
    def __init__(
        self,
        origin_node: Node,
        destination_node: Node,
        is_symmetric: bool = True,
        **kwargs,
    ):
        """
        Initializes an Arc entity connecting two nodes with processing parameters and metadata.

        Arcs represent connections between nodes in the simulation, modeling the flow of entities such as orders or goods.

        Required Arguments:

        - origin_node (Node): The node where the arc originates.
        - destination_node (Node): The node where the arc terminates.
        - kwargs: Additional keyword arguments passed to the SimulationEntity __init__ function.

        Optional Arguments:

        - is_symmetric (bool): Indicates whether the arc is symmetric, meaning the processing parameters apply equally in both directions.
            - Default: True
        """
        super().__init__(**kwargs)
        self.is_symmetric = is_symmetric
        self.origin_node = origin_node
        self.destination_node = destination_node



def hard_round(value: float, decimals: int = 2) -> float:
    """
    Rounds a float to a specified number of decimal places with hard rounding.

    Required Arguments:

    - value (float): The float value to be rounded.

    Optional Arguments:

    - decimals (int): The number of decimal places to round to.
        - Default: 2

    Returns:

    - float: The rounded float value.
    """
    factor = 10**decimals
    return round(value * factor) / factor



from scar_sim.entity import Arc, Node, SimulationObject


class Order(SimulationObject):
    def __init__(
        self,
        origin_node: Node,
        destination_node: Node,
        units: int,
        planned_path: list[int],
    ):
        """
        Initializes an Order object representing a shipment from an origin to a destination.

        Required Arguments:

        - origin_node (Node): The starting node for the order.
        - destination_node (Node): The ending node for the order.
        - units (int): The number of units in the order.
        - planned_path (list[int]): The planned path for the order as a list of graph IDs.
        """
        super().__init__()
        self.origin_node = origin_node
        self.destination_node = destination_node
        self.units = units

        # History of the Order's progress
        self.history = []

        # Simulation and miscellaneous state
        self.__simulation__ = None
        self.__current_object__ = self.origin_node
        self.__started__ = False
        self.__prev_time__ = 0.0

        # Note: Planned path is in terms of graph IDs and not in terms of simulation ids
        self.__set_planned_path__(planned_path, initial=True)
        self.__current_path_idx__ = 0

    def start(self) -> None:
        """
        Starts the Order's processing within the simulation.

        Raises an error if the Order has already been started.

        This method schedules the first event in the Order's lifecycle, transitioning it to the "started" status.

        Returns:

        - None
        """
        if self.__started__:
            raise ValueError("Order has already been started")
        self.__started__ = True
        self.__prev_time__ = self.__simulation__.current_time()
        self.__next__(status="started")

    def set_current_cashflow(self, cashflow: int | float) -> None:
        """
        Sets the cashflow for the most recent history entry of the Order.

        While designed to be used internally, this method can be overridden for custom cashflow handling.

        Required Arguments:

        - cashflow (int | float): The cashflow amount to set for the current history entry.
        """
        self.history[-1]["cashflow"] = cashflow

    def inject_metadata(self) -> dict:
        """
        Inject metadata about the Order into the history log entry when called. This is called by the __next__ method without arguments when logging history.

        This method can be overridden to provide custom metadata such as time, order ID, customer info, product info, etc.

        By default, this returns the current simulation time in integer form as 'time'.
        """
        return {
            "time": int(self.__simulation__.current_time()),
        }

    def consider_reroute(self) -> tuple[bool, list[int]]:
        """
        Consider whether to reroute the Order at its current location.

        This method can be overridden to provide custom rerouting logic based on current simulation state, congestion, delays, etc.

        This method will not have any inputs, but can access the Order's current state.

        By default, this placeholder does not reroute and returns (False, None).
        """
        return (False, list())

    def __set_planned_path__(
        self, planned_path: list[int], initial=False
    ) -> None:
        """
        An internal method to set the planned path for the Order.

        This is expected to be called internally when the Order is created and when rerouting is performed.

        Required Arguments:

        - planned_path (list[int]): The new planned path as a list of graph IDs.
        - initial (bool): Whether this is the initial setting of the planned path. Defaults to False.

        Raises:

        - ValueError: If the new path does not match the current path up to the current location (when not initial).
        - AssertionError: If the planned path does not start at the origin node or end at the destination node.
            - Note: More specifically, the planned path must start with the origin node's inbound graph ID and end with the destination node's inbound graph ID.

        Returns:
        - None
        """
        if not initial:
            if (
                planned_path[: self.__current_path_idx__ + 1]
                != self.__planned_path__[: self.__current_path_idx__ + 1]
            ):
                raise ValueError(
                    "New path must match current path up to current location"
                )
        assert (
            planned_path[0] == self.origin_node.inbound_graph_id
        ), "Planned path must start at origin node"
        assert (
            planned_path[-1] == self.destination_node.inbound_graph_id
        ), "Planned path must end at destination node"
        self.__planned_path__ = planned_path

    def __next__(self, status: str) -> None:
        """
        An internal method to progress the Order to the next status in its lifecycle.

        This method is expected to be stored as an event in the simulation's event queue and fired off at the appropriate times.

        Required Arguments:

        - status (str): The current status of the Order. Must be one of "started", "shipped", "arrived", or "completed".

        Raises:

        - ValueError: If the status is unknown or if the current object does not match the expected type for the status transition.

        Returns:

        - None
        """
        # Log this item into the Order history
        self.history.append(
            {
                "time": self.__simulation__.current_time(),
                "time_delta": round(
                    self.__simulation__.current_time() - self.__prev_time__, 3
                ),
                "order_id": self.id,
                "current_obj_id": self.__current_object__.id,
                "meta": self.__current_object__.get_metadata(
                    **self.inject_metadata()
                ),
                "status": status,
                "cashflow": 0.0,
            }
        )
        self.__current_path_idx__ += 1
        self.__prev_time__ = self.__simulation__.current_time()

        if status == "started":
            # Validate that we are at a Node that can process orders
            if not isinstance(self.__current_object__, Node):
                raise ValueError("Current object must be a Node when started")
            # Perform any logic at the origin node to process the order (i.e., remove from capacity)
            self.__current_object__.order_placed(self)
            # Set up for shipping
            next_status = "shipped"
        elif status == "shipped":
            # If the current object is a Node, we need to ship the order (i.e., remove from inventory)
            if isinstance(self.__current_object__, Node):
                self.__current_object__.order_shipped(self)

            # Pay for processing an order when it is shipped from a node
            self.set_current_cashflow(
                self.__current_object__.get_cashflow(units=self.units)
            )
            # Given the planned path, set the current object to the next planned Arc
            next_graph_id = self.__planned_path__[self.__current_path_idx__]
            self.__current_object__ = self.__simulation__.graph.arc_obj_graph[
                self.__current_object__.outbound_graph_id
            ][next_graph_id]
            # Determine the next node given this arc (allowing for symmetic arcs)
            if (
                self.__current_object__.origin_node.inbound_graph_id
                == next_graph_id
            ):
                self.__next_node__ = self.__current_object__.origin_node
            else:
                self.__next_node__ = self.__current_object__.destination_node
            next_status = "arrived"
        elif status == "arrived":
            if not isinstance(self.__current_object__, Arc):
                raise ValueError("Current object must be an Arc when arriving")
            # Pay for the transportation when a unit arrives at the destination node
            self.set_current_cashflow(
                self.__current_object__.get_cashflow(units=self.units)
            )
            # Set the current object to the next Node (supports symmetric arcs)
            self.__current_object__ = self.__next_node__
            if not isinstance(self.__current_object__, Node):
                raise ValueError("Next object must be a Node when arriving")
            # Consider rerouting the order
            should_reroute, new_path = self.consider_reroute()
            if should_reroute:
                self.__set_planned_path__(new_path)
            if isinstance(self.__current_object__, Node):
                # Only fire off the order arrived event if we are at a Node and a reroute has not changed the path away from this Node
                # Fire off the order arrived event at the Node for processing (i.e., add to inventory)
                if len(self.__planned_path__) == self.__current_path_idx__:
                    # We are at the end of the path
                    self.__current_object__.order_arrived(self)
                elif (
                    self.__planned_path__[self.__current_path_idx__ + 1]
                    == self.__current_object__.outbound_graph_id
                ):
                    self.__current_object__.order_arrived(self)
            next_status = (
                "completed"
                if self.__current_object__.id == self.destination_node.id
                else "shipped"
            )
        elif status == "completed":
            # Validate that we are at a Node that can receive Orders
            if not isinstance(self.__current_object__, Node):
                raise ValueError("Current object must be a Node when completed")
            # Fire off the order completed event at the Node for processing (i.e., add to capacity)
            self.__current_object__.order_completed(self)
            # When called with "completed", we do not schedule any further events
            return
        else:
            raise ValueError(f"Unknown status: {status}")

        self.__simulation__.add_event(
            time_delta=(
                self.__current_object__.get_processing_time()
                if next_status != "completed"
                else 0.0
            ),
            func=self.__next__,
            kwargs={"status": next_status},
        )



from scar_sim.queue import Queue
from scar_sim.entity import Node, Arc, SimulationObject
from scar_sim.order import Order
from scar_sim.graph import Graph
import dill


class Simulation:
    def __init__(self):
        """
        Initializes a Simulation object to manage the overall simulation state, including objects, orders, event queue, and graph representations.
        """
        # Simulation objects
        self.objects = []
        self.orders = []

        # Stateful queue and graphs
        self.__queue__ = Queue()
        self.graph = Graph()

    def current_time(self) -> float:
        """
        Returns the current simulation time.

        Returns:

        - float: The current time in the simulation.
        """
        return self.__queue__.__current_time__

    def add_event(
        self,
        time_delta: float,
        func,
        args: tuple = tuple(),
        kwargs: dict = dict(),
    ) -> None:
        """
        Schedules a new event in the simulation's event queue.

        Required Arguments:

        - time_delta (float): The time delay after which the event should be executed. Must be non-negative.
        - func (callable): The function to be called when the event is processed.

        Optional Arguments:

        - args (tuple): Positional arguments to pass to the function when called.
            - Default: tuple()
        - kwargs (dict): Keyword arguments to pass to the function when called.
            - Default: dict()

        Returns:

        - None
        """
        self.__queue__.add(
            time_delta=time_delta, func=func, args=args, kwargs=kwargs
        )

    def add_object(self, obj: Node | Arc | Order) -> SimulationObject:
        """
        Adds a SimulationObject (Node, Arc, or Order) to the simulation.
        Assigns a unique ID to the object and updates relevant simulation structures.

        Required Arguments:

        - obj (Node | Arc | Order): The object to be added to the simulation.

        Raises:

        - ValueError: If the object is already part of a simulation.
        - ValueError: If the object type is not recognized (not Node, Arc, or Order).

        Returns:

        - SimulationObject: The added object with updated simulation references and IDs.
        """
        if obj.id is not None:
            raise ValueError("Object is already added to a simulation")
        if isinstance(obj, (Node, Arc, Order)):
            # Create a ref to this simulation in the object
            obj.__simulation__ = self
            # Set the object ID such that it is unique within this simulation
            obj.id = len(self.objects)
            self.objects.append(obj)
            if isinstance(obj, (Node, Arc)):
                # Add arcs and nodes to the graph and update graph structures
                self.graph.add_object(obj)
            elif isinstance(obj, Order):
                # Add orders to the order list and set an order_id specific to the stored order list
                obj.order_id = len(self.orders)
                self.orders.append(obj)
        else:
            raise ValueError("Object type not recognized for simulation")
        return obj

    def run(self, max_time: float):
        """
        Runs the simulation until the specified maximum time is reached.

        Required Arguments:

        - max_time (float): The maximum simulation time to run until.

        Returns:

        - None
        """
        self.__queue__.run(max_time=max_time)

    def export_state(self, filename: str | None = None) -> bytes | str:
        """
        Exports the current state of the simulation using dill serialization.

        Optional Arguments:

        - filename (str | None): The file path to save the serialized state. If None, the state is returned as bytes.
            - Default: None

        Returns:

        - bytes | str: The serialized state as bytes if filename is None, otherwise the filename where the state was saved.
        """
        if filename is not None:
            with open(filename, "wb") as f:
                dill.dump(self, f)
            return filename
        else:
            return dill.dumps(self)

    @staticmethod
    def import_state(
        data: bytes | None = None, filename: str | None = None
    ) -> "Simulation":
        """
        Imports a simulation state from dill serialized data or a file.
        If both data and filename are provided, the filename takes precedence.

        Required Arguments (one of):

        - data (bytes | None): The serialized simulation state as bytes.
        - filename (str | None): The file path to load the serialized state from.

        Raises:

        - ValueError: If neither data nor filename is provided.

        Returns:

        - Simulation: The imported Simulation object.

        """
        if filename is not None:
            with open(filename, "rb") as f:
                return dill.load(f)
        elif data is not None:
            return dill.loads(data)
        else:
            raise ValueError(
                "Either data or filename must be provided to import a simulation"
            )



from scar_sim.entity import Node, Arc
from scgraph import Graph as SCGraph
from typing import Literal


class Graph:
    def __init__(self):
        """
        Initializes a Graph object to manage nodes and arcs within the simulation.
        """
        self.time_graph = []
        """Initializes the time graph as a list of dictionaries. This maps index based graph IDs to time values."""
        self.cashflow_graph = []
        """Initializes the cashflow graph as a list of dictionaries. This maps index based graph IDs to cashflow values. Cashflows are stored as negative values since we are trying to minimize costs (which are negative cashflows)."""
        self.arc_obj_graph = []
        """Initializes the arc object graph as a list of dictionaries. This maps index based graph IDs to Arc objects."""

    def update_graphs(self, obj: Arc | Node):
        """
        Updates the graph representations with the provided Arc or Node object.

        This should be called when an Arc or Node's processing parameters are changed to ensure the graph representations remain accurate.

        Any cashflow that is positive will be treated as zero since cashflows in the graph are only used to represent costs (negative cashflows).

        If an arc is symmetric, the graph will be updated in both directions.

        Required Arguments:

        - obj (Arc | Node): The Arc or Node object to update the graphs with.

        Returns:

        - None
        """
        if isinstance(obj, Arc):
            self.time_graph[obj.origin_node.outbound_graph_id][
                obj.destination_node.inbound_graph_id
            ] = max(float(obj.processing_time_avg), 0)
            self.cashflow_graph[obj.origin_node.outbound_graph_id][
                obj.destination_node.inbound_graph_id
            ] = max(-float(obj.processing_cashflow_per_unit), 0)
            self.arc_obj_graph[obj.origin_node.outbound_graph_id][
                obj.destination_node.inbound_graph_id
            ] = obj
            if obj.is_symmetric:
                self.time_graph[obj.destination_node.outbound_graph_id][
                    obj.origin_node.inbound_graph_id
                ] = max(float(obj.processing_time_avg), 0)
                self.cashflow_graph[obj.destination_node.outbound_graph_id][
                    obj.origin_node.inbound_graph_id
                ] = max(-float(obj.processing_cashflow_per_unit), 0)
                self.arc_obj_graph[obj.destination_node.outbound_graph_id][
                    obj.origin_node.inbound_graph_id
                ] = obj
        elif isinstance(obj, Node):
            self.time_graph[obj.inbound_graph_id] = {
                obj.outbound_graph_id: max(float(obj.processing_time_avg), 0)
            }
            self.cashflow_graph[obj.inbound_graph_id] = {
                obj.outbound_graph_id: max(
                    -float(obj.processing_cashflow_per_unit), 0
                )
            }

    def add_object(self, obj: Node | Arc) -> Node | Arc:
        """
        Adds a Node or Arc object to the graph representations.

        Nodes are actually represented as two nodes in the graph (inbound and outbound) connected by an arc that models processing time and cashflow.

        Graphs are updated accordingly.

        Required Arguments:

        - obj (Node | Arc): The Node or Arc object to add to the graphs.
        """
        if isinstance(obj, Node):
            obj.inbound_graph_id = len(self.time_graph)
            obj.outbound_graph_id = obj.inbound_graph_id + 1
            self.time_graph += [dict(), dict()]
            self.cashflow_graph += [dict(), dict()]
            self.arc_obj_graph += [dict(), dict()]
        if isinstance(obj, Node | Arc):
            self.update_graphs(obj)
        return obj

    def get_path_weight(
        self, path: list[int], graph: Literal["cashflow", "time"]
    ) -> float:
        """
        Given a path (list of graph IDs), calculates the total weight of the path based on the specified graph type.

        Required Arguments:

        - path (list[int]): A list of graph IDs representing the path.
        - graph (Literal['cashflow', 'time']): The type of graph to use for weight calculation.

        Returns:

        - float: The total weight of the path.
            - Note: Since cashflows are stored as negative values, the returned cashflow weight will be adjusted back to its original sign for consistency.
        """
        graph_obj = (
            self.cashflow_graph if graph == "cashflow" else self.time_graph
        )
        weight_sum = 0.0
        for i in range(len(path) - 1):
            origin = path[i]
            destination = path[i + 1]
            weight_sum += graph_obj[origin][destination]
        return weight_sum if graph == "time" else -weight_sum

    def get_optimal_path(
        self,
        origin_node: Node,
        destination_node: Node,
        graph: Literal["cashflow", "time"],
    ) -> list[int]:
        """
        Calculates the optimal path between the specified origin and destination nodes based on the specified graph type.

        Required Arguments:

        - origin_node (Node): The starting node for the path.
        - destination_node (Node): The ending node for the path.
        - graph (Literal['cashflow', 'time']): The type of graph to use for path calculation.

        Returns:

        - list[int]: A list of graph IDs representing the optimal path.
        """
        graph_obj = (
            self.cashflow_graph if graph == "cashflow" else self.time_graph
        )
        return SCGraph.dijkstra_makowski(
            graph_obj,
            origin_node.inbound_graph_id,
            destination_node.inbound_graph_id,
        )["path"]

    def get_route_options(
        self, origin_node: Node, destination_node: Node
    ) -> dict:
        """
        Calculates route options between the specified origin and destination nodes.

        Required Arguments:

        - origin_node (Node): The starting node for the route.
        - destination_node (Node): The ending node for the route.

        Returns:

        - dict: A dictionary containing route options for minimum cashflow and minimum time paths, each with their respective path, time, and cashflow values.
            - Example:
            ```
            {
                "min_cashflow":{
                    'path': [...],
                    'time': ...,
                    'cashflow': ...
                },
                "min_time":{
                    'path': [...],
                    'time': ...,
                    'cashflow': ...
                }
            }
            ```
        """
        min_time_path = self.get_optimal_path(
            origin_node, destination_node, "time"
        )
        min_cashflow_path = self.get_optimal_path(
            origin_node, destination_node, "cashflow"
        )
        return {
            "min_cashflow": {
                "path": min_cashflow_path,
                "time": self.get_path_weight(min_cashflow_path, "time"),
                "cashflow": self.get_path_weight(min_cashflow_path, "cashflow"),
            },
            "min_time": {
                "path": min_time_path,
                "time": self.get_path_weight(min_time_path, "time"),
                "cashflow": self.get_path_weight(min_time_path, "cashflow"),
            },
        }
